# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
# ruff: noqa: E402, F401

import importlib.metadata

from . import nexus, uncertainty, time_of_flight

try:
    __version__ = importlib.metadata.version("essreduce")
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

del importlib

__all__ = ["nexus", "uncertainty", "time_of_flight"]
